using System;
using System.Net;
using System.Net.Sockets;
using System.Text;
using UnityEngine;

// State object for reading client data asynchronously
public class StateObject
{
    // Client socket.
    public Socket workSocket = null;
    // Size of receive buffer.
    public const int BufferSize = 1024;
    // Receive buffer.
    public byte[] buffer = new byte[BufferSize];
    // Received data string.
    public StringBuilder sb = new StringBuilder();
}

public class AsynchronousSocketCommunicationManager
{
    private static Socket staticHandler = null;

    private static String socketIpAddress = "127.0.0.1";
    private static int socketPort = 11111;
    private static Encoding socketEncoding = Encoding.ASCII;
    private static String socketEOCString = "<EOC>";

    private static bool connected = false;
    private static bool readed = false;
    private static bool sended = false;

    private static bool pendingListeningRequest = false;
    private static bool pendingReadingRequest = false;
    private static bool pendingSendingRequest = false;

    private static String content = String.Empty;

    public AsynchronousSocketCommunicationManager()
    { }

    public static bool IsConnected()
    {
        return connected;
    }

    public static bool IsReaded()
    {
        return readed;
    }

    public static bool IsSended()
    {
        return sended;
    }

    public static String GetContent()
    {
        return content;
    }

    public static bool Listen(String ipAddress, int port, Encoding encoding, String eocString)
    {
        if (!pendingListeningRequest && staticHandler == null)
        {
            ResetStatus();

            socketIpAddress = (ipAddress != null && ipAddress != String.Empty) ? ipAddress : socketIpAddress;
            socketPort = (port >= 10000) ? port : socketPort;
            socketEncoding = (encoding != null) ? encoding : socketEncoding;
            socketEOCString = (eocString != null && eocString != String.Empty) ? eocString : socketEOCString;

            IPEndPoint localEndPoint = new IPEndPoint(IPAddress.Parse(socketIpAddress), socketPort);
            Socket listener = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);

            try
            {
                Debug.Log("[AsynchronousSocketCommunicationManager]: Waiting connection...");

                connected = false;

                listener.Bind(localEndPoint);
                listener.Listen(100);

                pendingListeningRequest = true;
                listener.BeginAccept(new AsyncCallback(AcceptCallback), listener);
            }
            catch (Exception e)
            {
                Debug.Log(e.ToString());
                pendingListeningRequest = false;
                return false;
            }

            return true;
        }
        else {
            return false;
        }
    }

    public static void AcceptCallback(IAsyncResult ar)
    {
        Socket listener = (Socket)ar.AsyncState;
        staticHandler = listener.EndAccept(ar);

        connected = true;
        pendingListeningRequest = false;

        Debug.Log("[AsynchronousSocketCommunicationManager]: Connection established!");
    }

    public static bool Read()
    {
        if (!pendingReadingRequest)
        {
            // Debug.Log("[AsynchronousSocketCommunicationManager]: Reading data...");

            readed = false;
            content = String.Empty;

            StateObject state = new StateObject();
            state.workSocket = staticHandler;

            pendingReadingRequest = true;
            staticHandler.BeginReceive(state.buffer, 0, StateObject.BufferSize, 0, new AsyncCallback(ReadCallback), state);

            return true;
        }
        else {
            return false;
        }

    }

    public static void ReadCallback(IAsyncResult ar)
    {
        StateObject state = (StateObject)ar.AsyncState;
        Socket handler = state.workSocket;

        int bytesRead = handler.EndReceive(ar);

        if (bytesRead > 0)
        {
            // There  might be more data, so store the data received so far.
            state.sb.Append(socketEncoding.GetString(state.buffer, 0, bytesRead));
            // Check for end-of-file tag. If it is not there, read more data.
            content = state.sb.ToString();
            if (content.IndexOf(socketEOCString) > -1)
            {
                // All the data has been read from the client.
                content = content.Substring(0, content.IndexOf(socketEOCString));

                readed = true;
                pendingReadingRequest = false;

                // Debug.Log("[AsynchronousSocketCommunicationManager]: Data readed!");
            }
            else {
                // Not all data received. Get more.
                handler.BeginReceive(state.buffer, 0, StateObject.BufferSize, 0, new AsyncCallback(ReadCallback), state);
            }
        }
    }

    public static bool Send(String data)
    {
        if (!pendingSendingRequest)
        {
            // Debug.Log("[AsynchronousSocketCommunicationManager]: Sending data...");

            sended = false;

            byte[] byteData = socketEncoding.GetBytes(String.Concat(data,socketEOCString));

            pendingSendingRequest = true;
            staticHandler.BeginSend(byteData, 0, byteData.Length, 0, new AsyncCallback(SendCallback), staticHandler);

            return true;
        }
        else {
            return false;
        }
    }

    public static void SendCallback(IAsyncResult ar)
    {
        try
        {
            Socket handler = (Socket)ar.AsyncState;
            int bytesSent = handler.EndSend(ar);

            sended = true;
            pendingSendingRequest = false;

            // Debug.Log("[AsynchronousSocketCommunicationManager]: Data sended!");
        }
        catch (Exception e)
        {
            Debug.Log(e.ToString());
        }
    }

    public static bool Close()
    {
        try
        {
            staticHandler.Shutdown(SocketShutdown.Both);
            staticHandler.Close();
            staticHandler = null;

            Debug.Log("[AsynchronousSocketCommunicationManager]: Connection closed.");
        }
        catch (Exception e)
        {
            Debug.Log(e.ToString());
            return false;
        }

        ResetStatus();

        return true;
    }

    private static void ResetStatus() {
        connected = false;
        readed = false;
        sended = false;

        pendingListeningRequest = false;
        pendingReadingRequest = false;
        pendingSendingRequest = false;
    }

}